/*
 * Copyright (c) 2019 Giam Teck Choon <choon@choon.net> or CHOON.NET.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <bglibs/str.h>
#include <bglibs/dns.h>

#include "mailfront.h"

#define elogf(A, ...) { if(A){ showpid(); fprintf(stderr, __VA_ARGS__); } }

static response resp;
static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-helo";
static pid_t pid;
static int is_ipv6 = 0;
static int debug = 0;
static int respnum = 0;
static const char *local_host = NULL;
static const char *local_ip = NULL;
static const char *remote_host = NULL;
static const char *remote_ip = NULL;
static const char *relayclient = NULL;
static const char *rejectcheck = NULL;
static const char *defercheck = NULL;
static const char *skipcheck = NULL;
static const char *addheader = NULL;
static const char *checkmatchip = NULL;
static const char *checkmatchipclass = NULL;
static str heloname;
static str estr;
static str hmsg;

void showpid()
{
	if (!pid || pid < 0) {
		pid = getpid();
	}
	fprintf(stderr, "%s-%s[%d]: ", mainprog, thisprog, pid);
}

/* Return 1 if is valid IPv4 or IPv6 otherwise return 0 */
int is_valid_ip(const char *ipstr)
{
	if (strchr(ipstr, ':')) {
		ipv6addr ip6;
		if (ipv6_scan(ipstr, &ip6)) {
			return 1;
		}
	} else {
		ipv4addr ip4;
		if (ipv4_scan(ipstr, &ip4)) {
			return 1;
		}
	}
	return 0;
}

/* Return 1 if within the match class otherwise return 0.
 * For IPv4, we lookup to see whether it is the same class C
 * or compare the first 3 bytes.
 * For IPv6, we lookup for the first 12 bytes whether it is the same
 */
int match_ipclass(const char *ipstr1, const char *ipstr2)
{
	if (strchr(ipstr1, ':') && strchr(ipstr2, ':')) {
		ipv6addr ip1;
		ipv6addr ip2;
		if (!ipv6_scan(ipstr1, &ip1)) {
			return 0;
		}
		if (!ipv6_scan(ipstr2, &ip2)) {
			return 0;
		}
		if (memcmp(ip1.addr, ip2.addr, 12) == 0) {
			return 1;
		}
	} else {
		ipv4addr ip1;
		ipv4addr ip2;
		if (!ipv4_scan(ipstr1, &ip1)) {
			return 0;
		}
		if (!ipv4_scan(ipstr2, &ip2)) {
			return 0;
		}
		if (memcmp(ip1.addr, ip2.addr, 3) == 0) {
			return 1;
		}
	}
	return 0;
}

int dns_A_record(struct dns_result *out, const char *fqdn)
{
	if (dns_ip4(out, fqdn) < 0) {
		return -1;
	}
	return out->count;
}

int dns_AAAA_record(struct dns_result *out, const char *fqdn)
{
	if (dns_ip6(out, fqdn) < 0) {
		return -1;
	}
	return out->count;
}

/* return 0 if not found char c in char *s otherwise return > 0
 * if respnum > 0, set resp.message and return 2
 */
int check_senderhelo_char_start(const char *s, char c)
{
	char d = '\'';
	if (c == '\'') {
		d = '"';
	}
	/* first character */
	if (*s == c) {
		if (respnum > 0) {
			str_init(&estr);
			str_cats(&estr, "Sorry, your HELO name starts with ");
			str_cats(&estr, "disallowed character ");
			str_catc(&estr, d);
			str_catc(&estr, c);
			str_catc(&estr, d);
			resp.message = estr.s;
			return 2;
		}
		elogf(debug, "%c%c%c found in first character of HELO name "
		      "%s\n", d, c, d, s);
		str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_catc(&hmsg, d);
		str_catc(&hmsg, c);
		str_catc(&hmsg, d);
		str_cats(&hmsg, " found in first character of HELO");
		str_cats(&hmsg, " name ");
		str_cats(&hmsg, s);
		str_cats(&hmsg, "\n");
		return 1;
	}
	return 0;
}

/* return 0 if not found char c in char *s otherwise return > 0
 * if respnum > 0, set resp.message and return 2
 */
int check_senderhelo_char_end(const char *s, char c)
{
	char d = '\'';
	if (c == '\'') {
		d = '"';
	}
	int len = strlen(s);
	/* last character */
	if (s[len-1] == c) {
		if (respnum > 0) {
			str_init(&estr);
			str_cats(&estr, "Sorry, your HELO name ends with ");
			str_cats(&estr, "disallowed character ");
			str_catc(&estr, d);
			str_catc(&estr, c);
			str_catc(&estr, d);
			resp.message = estr.s;
			return 2;
		}
		elogf(debug, "%c%c%c found in last character of HELO name"
		      " %s\n", d, c, d, s);
		str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_catc(&hmsg, d);
		str_catc(&hmsg, c);
		str_catc(&hmsg, d);
		str_cats(&hmsg, " found in last character of HELO");
		str_cats(&hmsg, " name ");
		str_cats(&hmsg, s);
		str_cats(&hmsg, "\n");
		return 1;
	}
	return 0;
}

/* return 0 if not found char c in char *s otherwise return > 0
 * if respnum > 0, set resp.message and return 2
 */
int check_senderhelo_char(const char *s, char c)
{
	if (strchr(s, c)) {
		char d = '\'';
		if (c == '\'') {
			d = '"';
		}
		if (respnum > 0) {
			str_init(&estr);
			str_cats(&estr, "Sorry, your HELO name has the ");
			str_cats(&estr, "disallowed character ");
			str_catc(&estr, d);
			str_catc(&estr, c);
			str_catc(&estr, d);
			resp.message = estr.s;
			return 2;
		}
		elogf(debug, "%c%c%c character found in HELO name %s\n",
		      d, c, d, s);
		str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_catc(&hmsg, d);
		str_catc(&hmsg, c);
		str_catc(&hmsg, d);
		str_cats(&hmsg, " character found in HELO name ");
		str_cats(&hmsg, s);
		str_cats(&hmsg, "\n");
		return 1;
	}
	return 0;
}

/* return 0 for checking PASS otherwise return > 0
 * if respnum > 0, set resp.message and return 2
 */
int check_heloname_ipv4()
{
	int i;
	struct dns_result a = {0};

	if (dns_A_record(&a, heloname.s)) {
		if (a.count > 0) {
			str ans = {0};
			str hdr = {0};
			int x = 0;
			int y = 0;
			for (i=0; i<a.count; i++) {
				str_copys(&ans, ipv4_format(&a.rr.ip4[i]));
				elogf(debug, "HELO name %s DNS lookup A record"
				      " %s\n", heloname.s, ans.s);
				if (strcmp(ans.s, remote_ip) == 0) {
					elogf(debug, "HELO name IPv4 %s match "
					      "REMOTEIP %s\n", ans.s,
					      remote_ip);
					x = 1;
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv4=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " match REMOTEIP=");
					str_cat2s(&hdr, remote_ip, "\n");
				} else {
					elogf(debug, "HELO name IPv4 %s "
					      "doesn't match REMOTEIP %s\n",
					      ans.s, remote_ip);
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv4=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " does't match ");
					str_cats(&hdr, "REMOTEIP=");
					str_cat2s(&hdr, remote_ip, "\n");
				}
				if (match_ipclass(ans.s, remote_ip)) {
					elogf(debug, "HELO name IPv4 %s is "
					      "within REMOTEIP %s IPv4 class."
					      "\n", ans.s, remote_ip);
					y = 1;
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv4=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " match REMOTEIP=");
					str_cats(&hdr, remote_ip);
					str_cats(&hdr, " IPv4 class.\n");
				} else {
					elogf(debug, "HELO name IPv4 %s is not "
					      "within REMOTEIP %s IPv4 "
					      "class.\n", ans.s, remote_ip);
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv4=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " not match REMOTEIP=");
					str_cats(&hdr, remote_ip);
					str_cats(&hdr, " IPv4 class.\n");
				}
			}
			if (ans.len > 0) {
				str_free(&ans);
			}
			if (checkmatchip && x == 0) {
				if (respnum > 0) {
					str_init(&estr);
					str_cats(&estr, "Sorry, your HELO ");
					str_cats(&estr, "name IPv4 doesn't ");
					str_cats(&estr, "resolve to match ");
					str_cats(&estr, "your IPv4.");
					resp.message = estr.s;
					if (hdr.len > 0) str_free(&hdr);
					return 2;
				}
				if (hdr.len > 0) {
					str_copys(&hmsg, "X-HELO-Check-");
					str_cats(&hmsg, "Result: FAILED\n");
					str_cats(&hmsg, hdr.s);
					str_free(&hdr);
					return 1;
				}
			}
			if (checkmatchipclass && y == 0) {
				if (respnum > 0) {
					str_init(&estr);
					str_cats(&estr, "Sorry, your HELO ");
					str_cats(&estr, "name IPv4 doesn't ");
					str_cats(&estr, "resolve to match ");
					str_cats(&estr, "your IPv4 class.");
					resp.message = estr.s;
					if (hdr.len > 0) str_free(&hdr);
					return 2;
				}
				if (hdr.len > 0) {
					str_copys(&hmsg, "X-HELO-Check-");
					str_cats(&hmsg, "Result: FAILED\n");
					str_cats(&hmsg, hdr.s);
					str_free(&hdr);
					return 1;
				}
			}
			if (hdr.len > 0) str_free(&hdr);
		} else {
			elogf(debug, "Sorry, your HELO name %s DNS A record "
			      "lookup failed.\n", heloname.s);
			if (respnum > 0) {
				str_init(&estr);
				str_cats(&estr, "Sorry, your HELO name ");
				str_cats(&estr, heloname.s);
				str_cats(&estr, " DNS A record lookup failed.");
				resp.message = estr.s;
				dns_result_free(&a);
				return 2;
			}
			str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
			str_cats(&hmsg, "X-HELO-Check-Summary: HELO name ");
			str_cats(&hmsg, heloname.s);
			str_cats(&hmsg, " DNS A record lookup failed.\n");
			dns_result_free(&a);
			return 1;
		}
		dns_result_free(&a);
	}
	return 0;
}

/* return 0 for checking PASS otherwise return > 0
 * if respnum > 0, set resp.message and return 2
 */
int check_heloname_ipv6()
{
	int i;
	struct dns_result a = {0};

	if (dns_AAAA_record(&a, heloname.s)) {
		if (a.count > 0) {
			str ans = {0};
			str hdr = {0};
			int x = 0;
			int y = 0;
			for (i=0; i<a.count; i++) {
				str_copys(&ans, ipv6_format(&a.rr.ip6[i]));
				elogf(debug, "HELO name %s DNS lookup AAAA "
				      "record %s\n", heloname.s, ans.s);
				if (strcmp(ans.s, remote_ip) == 0) {
					elogf(debug, "HELO name IPv6 %s match "
					      "REMOTEIP %s\n", ans.s,
					      remote_ip);
					x = 1;
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv6=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " match REMOTEIP=");
					str_cat2s(&hdr, remote_ip, "\n");
				} else {
					elogf(debug, "HELO name IPv6 %s "
					      "doesn't match REMOTEIP %s\n",
					      ans.s, remote_ip);
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv6=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " does't match ");
					str_cats(&hdr, "REMOTEIP=");
					str_cat2s(&hdr, remote_ip, "\n");
				}
				if (match_ipclass(ans.s, remote_ip)) {
					elogf(debug, "HELO name IPv6 %s is "
					      "within REMOTEIP %s IPv6 class."
					      "\n", ans.s, remote_ip);
					y = 1;
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv6=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " match REMOTEIP=");
					str_cats(&hdr, remote_ip);
					str_cats(&hdr, " IPv6 class.\n");
				} else {
					elogf(debug, "HELO name IPv6 %s is not "
					      "within REMOTEIP %s IPv6 "
					      "class.\n", ans.s, remote_ip);
					str_cats(&hdr, "X-HELO-Check-Summary");
					str_cats(&hdr, ": HELO name IPv6=");
					str_cats(&hdr, ans.s);
					str_cats(&hdr, " not match REMOTEIP=");
					str_cats(&hdr, remote_ip);
					str_cats(&hdr, " IPv6 class.\n");
				}
			}
			if (ans.len > 0) {
				str_free(&ans);
			}
			if (checkmatchip && x == 0) {
				if (respnum > 0) {
					str_init(&estr);
					str_cats(&estr, "Sorry, your HELO ");
					str_cats(&estr, "name IPv6 doesn't ");
					str_cats(&estr, "resolve to match ");
					str_cats(&estr, "your IPv6.");
					resp.message = estr.s;
					if (hdr.len > 0) str_free(&hdr);
					return 2;
				}
				if (hdr.len > 0) {
					str_copys(&hmsg, "X-HELO-Check-");
					str_cats(&hmsg, "Result: FAILED\n");
					str_cats(&hmsg, hdr.s);
					str_free(&hdr);
					return 1;
				}
			}
			if (checkmatchipclass && y == 0) {
				if (respnum > 0) {
					str_init(&estr);
					str_cats(&estr, "Sorry, your HELO ");
					str_cats(&estr, "name IPv6 doesn't ");
					str_cats(&estr, "resolve to match ");
					str_cats(&estr, "your IPv6 class.");
					resp.message = estr.s;
					if (hdr.len > 0) str_free(&hdr);
					return 2;
				}
				if (hdr.len > 0) {
					str_copys(&hmsg, "X-HELO-Check-");
					str_cats(&hmsg, "Result: FAILED\n");
					str_cats(&hmsg, hdr.s);
					str_free(&hdr);
					return 1;
				}
			}
			if (hdr.len > 0) str_free(&hdr);
		} else {
			elogf(debug, "Sorry, your HELO name %s DNS AAAA "
			      "record lookup failed.\n", heloname.s);
			if (respnum > 0) {
				str_init(&estr);
				str_cats(&estr, "Sorry, your HELO name ");
				str_cats(&estr, heloname.s);
				str_cats(&estr, " DNS AAAA record lookup ");
				str_cats(&estr, "failed.");
				resp.message = estr.s;
				dns_result_free(&a);
				return 2;
			}
			str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
			str_cats(&hmsg, "X-HELO-Check-Summary: HELO name ");
			str_cats(&hmsg, heloname.s);
			str_cats(&hmsg, " DNS AAAA record lookup failed.\n");
			dns_result_free(&a);
			return 1;
		}
		dns_result_free(&a);
	}
	return 0;
}

static const response *init(void)
{
	local_ip = getprotoenv("LOCALIP");
	remote_ip = getprotoenv("REMOTEIP");
	local_host = getprotoenv("LOCALHOST");
	remote_host = getprotoenv("REMOTEHOST");
	if (remote_ip && *remote_ip) {
		if (strchr(remote_ip, ':')) {
			is_ipv6 = 1;
		}
	}
	relayclient = session_getenv("RELAYCLIENT");
	rejectcheck = session_getenv("HELO_CHECK_REJECT");
	defercheck = session_getenv("HELO_CHECK_DEFER");
	skipcheck = session_getenv("HELO_CHECK_SKIP");
	if (rejectcheck || defercheck) {
		respnum = 451;
		if (rejectcheck) {
			respnum = 553;
		}
		resp.number = respnum;
	}
	checkmatchip = session_getenv("HELO_CHECK_MATCH_IP");
	checkmatchipclass = session_getenv("HELO_CHECK_MATCH_IP_CLASS");
	if (session_getenv("HELO_CHECK_VERBOSE")) {
		debug = 1;
	}
	addheader = session_getenv("HELO_CHECK_ADD_HEADER");
	str_init(&heloname);
	return 0;
}

static const response *helo(str *hostname, str *capabilities)
{
	str_copys(&heloname, hostname->s);
	return 0;
	(void)hostname;
	(void)capabilities;
}

static const response *sender(str* address, str* params)
{
	str_init(&hmsg);
	/* sequence (1) */
	if (skipcheck) {
		elogf(debug, "HELO_CHECK_SKIP environment is set thus this "
		      "plugin skipped\n");
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: HELO_CHECK_SKIP ");
		str_cats(&hmsg, "is set\n");
		return 0;
	}
	if (!local_ip) {
		elogf(debug, "LOCALIP not found thus this plugin "
		      "skipped\n");
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: LOCALIP not set\n");
		return 0;
	}
	if (!remote_ip) {
		elogf(debug, "REMOTEIP not found thus this plugin "
		      "skipped\n");
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP not set\n");
		return 0;
	}
	/* sequence (2) */
	/* X-HELO-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4 */
	if (relayclient) {
		elogf(debug, "$RELAYCLIENT is set thus skipped for "
		      "REMOTEIP %s\n", remote_ip);
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_cat3s(&hmsg, "$RELAYCLIENT is set for ", remote_ip, "\n");
		return 0;
	}
	/* sequence (3) */
	/* X-HELO-Check-Result: SKIPPED: authenticated user from 1.2.3.4 */
	if (session_getnum("authenticated", 0)) {
		elogf(debug, "authenticated user thus skipped for "
		      "REMOTEIP %s\n", remote_ip);
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_cat3s(&hmsg, "authenticated user from ", remote_ip, "\n");
		return 0;
	}
	/* X-HELO-Check-Result: SKIPPED: LOCALIP is the same as REMOTEIP 1.2.3.4 */
	if (strcmp(local_ip, remote_ip) == 0) {
		elogf(debug, "LOCALIP=REMOTEIP=%s thus skipped\n",
		      remote_ip);
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_cats(&hmsg, "LOCALIP is the same as REMOTEIP ");
		str_cat2s(&hmsg, remote_ip, "\n");
		return 0;
	}
	/* sequence (4) */
	/* X-HELO-Check-Result: SKIPPED: $RBLSMTPD is set and empty for 1.1.1.1 */
	const char *rblsmtpd = getenv("RBLSMTPD");
	if (rblsmtpd && *rblsmtpd == 0) {
		elogf(debug, "$RBLSMTPD is set and empty thus skipped "
		      "for REMOTEIP %s\n", remote_ip);
		str_copys(&hmsg, "X-HELO-Check-Result: SKIPPED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_cats(&hmsg, "$RBLSMTPD is set and empty for ");
		str_cat2s(&hmsg, remote_ip, "\n");
		return 0;
	}
	/* sequence (5) */
	/* If heloname is empty */
	if (heloname.len == 0) {
		if (respnum > 0) {
			resp.message = "Sorry, your HELO name is empty!";
			return &resp;
		}
		elogf(debug, "helo name is empty!\n");
		str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: HELO name is empty!\n");
		return 0;
	}
	/* 0 continue, 1 to return 0 and 2 to return &resp; */
	int i, rc = 0;
	char checkstart[] = {'.', '[', '-', '\0'};
	/* sequence (5.1) */
	/* Don't accept start with a dot, [ or dash */
	for (i=0; i<strlen(checkstart); i++) {
		rc = check_senderhelo_char_start(heloname.s, checkstart[i]);
		if (rc == 2) return &resp;
		if (rc == 1) return 0;
	}
	/* sequence (5.2) */
	/* Don't accept end with a dot, ] and dash */
	char checkend[] = {'.', ']', '-', '\0'};
	for (i=0; i<strlen(checkend); i++) {
		rc = check_senderhelo_char_end(heloname.s, checkend[i]);
		if (rc == 2) return &resp;
		if (rc == 1) return 0;
	}
	/* sequence (5.3) */
	/* Must have at least a "." (dot) */
	if (strchr(heloname.s, '.') == NULL) {
		if (respnum > 0) {
			str_init(&estr);
			str_cats(&estr, "Sorry, your HELO name ");
			str_cats(&estr, "doesn't contains the ");
			str_cats(&estr, "character '.'");
			resp.message = estr.s;
			return &resp;
		}
		elogf(debug, "'.' character not found in "
		      "HELO name %s for REMOTEIP %s\n",
		      heloname.s, remote_ip);
		str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_cats(&hmsg, "'.' character not found ");
		str_cats(&hmsg, "in HELO name ");
		str_cats(&hmsg, heloname.s);
		str_cats(&hmsg, "\n");
		return 0;
	}
	/* sequence (5.4) */
	/* Disallowed character SPACE, TAB and !@#$%^&*()[]{}:;"'<>?/|\+=_, */
	char checks[] = {'!', '@', '#', '$', '%', '^', '&', '*', '(',
			 ')', '[', ']', '{', '}', ':', ';', '"', '\'',
			 '<', '>', '?', '/', '|', '\\', '+', '=',
			 '_', ' ', '\t', ',', '\0'};
	for (i=0; i<strlen(checks); i++) {
		rc = check_senderhelo_char(heloname.s, checks[i]);
		if (rc == 2) return &resp;
		if (rc == 1) return 0;
	}
	/* sequence (5.5) */
	/* Cannot use a valid IPv4 or IPv6 address */
	if (is_valid_ip(heloname.s)) {
		if (respnum > 0) {
			str_init(&estr);
			str_cats(&estr, "Sorry, your HELO name ");
			str_cats(&estr, "is a valid IP address ");
			str_cats(&estr, "which is disallowed");
			resp.message = estr.s;
			return &resp;
		}
		elogf(debug, "valid IP address found in "
		      "HELO name %s for REMOTEIP %s\n",
		      heloname.s, remote_ip);
		str_copys(&hmsg, "X-HELO-Check-Result: FAILED\n");
		str_cats(&hmsg, "X-HELO-Check-Summary: ");
		str_cats(&hmsg, "Valid IP address found ");
		str_cats(&hmsg, "in HELO name ");
		str_cats(&hmsg, heloname.s);
		str_cats(&hmsg, "\n");
		return 0;
	}
	/* sequence (6) */
	if (is_ipv6) {
		rc = check_heloname_ipv6();
		if (rc == 2) return &resp;
		if (rc == 1) return 0;
	} else {
		rc = check_heloname_ipv4();
		if (rc == 2) return &resp;
		if (rc == 1) return 0;
	}
	str_copys(&hmsg, "X-HELO-Check-Result: PASSED\n");
	str_cats(&hmsg, "X-HELO-Check-Summary: ");
	str_cat2s(&hmsg, "HELO name ", heloname.s);
	str_cat2s(&hmsg, " from REMOTEIP ", remote_ip);
	str_cats(&hmsg, " is valid.\n");
	return 0;
	(void)address;
	(void)params;
}

static const response *data_start(int fd)
{
	if (skipcheck) return 0;
	if (hmsg.len > 0 && addheader) {
		return backend_data_block(hmsg.s, hmsg.len);
	}
	str_init(&hmsg);
	return 0;
	(void)fd;
}

struct plugin plugin = {
	.version = PLUGIN_VERSION,
	.flags = 0,
	.init = init,
	.helo = helo,
	.sender = sender,
	.data_start = data_start,
};
